package org.dynmap.factions.runnable;

import com.massivecraft.factions.entity.BoardColl;
import com.massivecraft.factions.entity.Faction;
import com.massivecraft.factions.entity.FactionColl;
import com.massivecraft.factions.entity.MFlag;
import com.massivecraft.factions.entity.MFlagColl;
import com.massivecraft.factions.entity.MPlayer;
import com.massivecraft.factions.entity.Warp;
import com.massivecraft.massivecore.ps.PS;
import com.massivecraft.massivecore.store.EntityInternalMap;
import org.bukkit.ChatColor;
import org.bukkit.configuration.file.FileConfiguration;
import org.dynmap.factions.DynmapFactionsPlugin;
import org.dynmap.factions.config.AreaStyle;
import org.dynmap.factions.container.FactionBlock;
import org.dynmap.factions.container.FactionBlocks;
import org.dynmap.factions.container.TileFlags;
import org.dynmap.factions.property.Direction;
import org.dynmap.markers.AreaMarker;
import org.dynmap.markers.Marker;
import org.dynmap.markers.MarkerIcon;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Deque;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class FactionsUpdate implements Runnable {

    private static final String DEF_INFO_WINDOW = "<div class=\"infoWindow\"><span style=\"font-size:120%;\">" +
            "%regionName%</span><br/>Description: <span style=\"font-weight:bold;\">%description%</span><br/>Leader: " +
            "<span style=\"font-weight:bold;\">%playerOwners%</span><br/>Members: <span style=\"font-weight:bold;\">" +
            "%playerMembers%</span><br />Flags<br /><span style=\"font-weight:bold;\">%flags%</span></div>";
    private final String infoWindow;
    private final DynmapFactionsPlugin dynmapFactionsPlugin;
    public boolean runonce;

    public FactionsUpdate(DynmapFactionsPlugin dynmapFactionsPlugin, FileConfiguration configuration) {
        this.dynmapFactionsPlugin = dynmapFactionsPlugin;
        infoWindow = configuration.getString("infoWindow", DEF_INFO_WINDOW);
    }

    @Override
    public void run() {
        if (dynmapFactionsPlugin.isStop()) {
            return;
        }

        updateFactions();
        if (!runonce) {
            dynmapFactionsPlugin.getServer().getScheduler().scheduleSyncDelayedTask(dynmapFactionsPlugin,
                    this, dynmapFactionsPlugin.getUpdatePeriod());
        } else if (dynmapFactionsPlugin.getPending() == this) {
            dynmapFactionsPlugin.setPending(null);
        }
    }


    /* Update Factions information */
    private void updateFactions() {
        Map<String, AreaMarker> newMap = new HashMap<>();
        Map<String, Marker> newMarkers = new HashMap<>();

        /* Parse into faction centric mapping, split by world */
        Map<String, FactionBlocks> blocksByFaction = new HashMap<>();

        FactionColl fc = FactionColl.get();
        Collection<Faction> factions = fc.getAll();
        for (Faction faction : factions) {
            Set<PS> chunks = BoardColl.get().getChunks(faction);
            String fid = fc.getUniverse() + "_" + faction.getId();
            FactionBlocks factionBlocks = blocksByFaction.get(fid); /* Look up faction */
            if (factionBlocks == null) {    /* Create faction block if first time */
                factionBlocks = new FactionBlocks();
                blocksByFaction.put(fid, factionBlocks);
            }

            for (PS cc : chunks) {
                String world = cc.getWorld();

                /* Get block set for given world */
                List<FactionBlock> blocks = factionBlocks.blocks().computeIfAbsent(world, k -> new LinkedList<>());
                FactionBlock fb = new FactionBlock(cc.getChunkX(), cc.getChunkZ());
                blocks.add(fb); /* Add to list */
            }
        }
        /* Loop through factions */
        for (Faction faction : factions) {
            String factionName = ChatColor.stripColor(faction.getName());
            String fid = fc.getUniverse() + "_" + faction.getId();
            FactionBlocks factionBLocks = blocksByFaction.get(fid); /* Look up faction */
            if (factionBLocks == null) {
                continue;
            }

            /* Loop through each world that faction has blocks on */
            for (Map.Entry<String, List<FactionBlock>> worldBlocks : factionBLocks.blocks().entrySet()) {
                handleFactionOnWorld(factionName, faction, worldBlocks.getKey(), worldBlocks.getValue(), newMap);
            }
            factionBLocks.blocks().clear();

            /* Now, add marker for home location */
            EntityInternalMap<Warp> warps = faction.getWarps();
            Warp home = null;
            for (String warp : warps.keySet()) {
                if (warp.equalsIgnoreCase("home")) {
                    home = warps.get(warp);
                }
            }
            if (home == null) {
                continue;
            }
            PS homeLocation = home.getLocation();
            if (homeLocation != null) {
                String markId = fc.getUniverse() + "_" + factionName + "__home";
                MarkerIcon ico = getMarkerIcon(factionName);
                if (ico != null) {
                    Marker homeMarker = dynmapFactionsPlugin.getFactionMarkers().remove(markId);
                    String lbl = factionName + " [home]";
                    if (homeMarker == null) {
                        homeMarker = dynmapFactionsPlugin.getSet().createMarker(markId, lbl, homeLocation.getWorld(),
                                homeLocation.getLocationX(), homeLocation.getLocationY(), homeLocation.getLocationZ(), ico, false);
                    } else {
                        homeMarker.setLocation(homeLocation.getWorld(), homeLocation.getLocationX(), homeLocation.getLocationY(), homeLocation.getLocationZ());
                        homeMarker.setLabel(lbl);   /* Update label */
                        homeMarker.setMarkerIcon(ico);
                    }
                    if (homeMarker != null) {
                        homeMarker.setDescription(formatInfoWindow(faction)); /* Set popup */
                        newMarkers.put(markId, homeMarker);
                    }
                }
            }
        }
        blocksByFaction.clear();

        /* Now, review old map - anything left is gone */
        for (AreaMarker oldMarker : dynmapFactionsPlugin.getFactionAreaMarkers().values()) {
            oldMarker.deleteMarker();
        }
        for (Marker oldMarker : dynmapFactionsPlugin.getFactionMarkers().values()) {
            oldMarker.deleteMarker();
        }
        /* And replace with new map */
        dynmapFactionsPlugin.setFactionAreaMarkers(newMap);
        dynmapFactionsPlugin.setFactionMarkers(newMarkers);

    }

    /* Handle specific faction on specific world */
    private void handleFactionOnWorld(@NotNull String factionName, @NotNull Faction faction, @NotNull String world,
                                      @NotNull List<FactionBlock> blocks, @NotNull Map<String, AreaMarker> newmap) {
        double[] x;
        double[] z;
        int poly_index = 0; /* Index of polygon for given faction */

        /* Build popup */
        String desc = formatInfoWindow(faction);

        /* Handle areas */
        if (!isVisible(factionName, world) || blocks.isEmpty()) {
            return;
        }

        LinkedList<FactionBlock> nodeValues = new LinkedList<>();
        TileFlags currentBlocks = new TileFlags();
        /* Loop through blocks: set flags on block maps */
        for (FactionBlock b : blocks) {
            currentBlocks.setFlag(b.x(), b.z(), true); /* Set flag for block */
            nodeValues.addLast(b);
        }
        /* Loop through until we don't find more areas */
        while (nodeValues != null) {
            LinkedList<FactionBlock> ourNodes = null;
            LinkedList<FactionBlock> newList = null;
            TileFlags ourBlocks = null;
            int minX = Integer.MAX_VALUE;
            int minZ = Integer.MAX_VALUE;
            for (FactionBlock node : nodeValues) {
                int nodeX = node.x();
                int nodeZ = node.z();
                /* If we need to start shape, and this block is not part of one yet */
                if ((ourBlocks == null) && currentBlocks.getFlag(nodeX, nodeZ)) {
                    ourBlocks = new TileFlags();  /* Create map for shape */
                    ourNodes = new LinkedList<>();
                    floodFillTarget(currentBlocks, ourBlocks, nodeX, nodeZ);   /* Copy shape */
                    ourNodes.add(node); /* Add it to our node list */
                    minX = nodeX;
                    minZ = nodeZ;
                }
                /* If shape found, and we're in it, add to our node list */
                else if ((ourBlocks != null) && ourBlocks.getFlag(nodeX, nodeZ)) {
                    ourNodes.add(node);
                    if (nodeX < minX) {
                        minX = nodeX;
                        minZ = nodeZ;
                    } else if ((nodeX == minX) && (nodeZ < minZ)) {
                        minZ = nodeZ;
                    }
                } else {  /* Else, keep it in the list for the next polygon */
                    if (newList == null) {
                        newList = new LinkedList<>();
                    }
                    newList.add(node);
                }
            }
            nodeValues = newList; /* Replace list (null if no more to process) */
            if (ourBlocks != null) {
                /* Trace outline of blocks - start from minx, min z going to x+ */
                int cur_x = minX;
                int cur_z = minZ;
                Direction dir = Direction.POSITIVE_X;
                List<int[]> lineList = new ArrayList<>();
                lineList.add(new int[]{minX, minZ}); // Add start point
                while ((cur_x != minX) || (cur_z != minZ) || (dir != Direction.NEGATIVE_Z)) {
                    switch (dir) {
                        case POSITIVE_X: /* Segment in X+ direction */
                            if (!ourBlocks.getFlag(cur_x + 1, cur_z)) { /* Right turn? */
                                lineList.add(new int[]{cur_x + 1, cur_z}); /* Finish line */
                                dir = Direction.POSITIVE_Z;  /* Change direction */
                            } else if (!ourBlocks.getFlag(cur_x + 1, cur_z - 1)) {  /* Straight? */
                                cur_x++;
                            } else {  /* Left turn */
                                lineList.add(new int[]{cur_x + 1, cur_z}); /* Finish line */
                                dir = Direction.NEGATIVE_Z;
                                cur_x++;
                                cur_z--;
                            }
                            break;
                        case POSITIVE_Z: /* Segment in Z+ direction */
                            if (!ourBlocks.getFlag(cur_x, cur_z + 1)) { /* Right turn? */
                                lineList.add(new int[]{cur_x + 1, cur_z + 1}); /* Finish line */
                                dir = Direction.NEGATIVE_X;  /* Change direction */
                            } else if (!ourBlocks.getFlag(cur_x + 1, cur_z + 1)) {  /* Straight? */
                                cur_z++;
                            } else {  /* Left turn */
                                lineList.add(new int[]{cur_x + 1, cur_z + 1}); /* Finish line */
                                dir = Direction.POSITIVE_X;
                                cur_x++;
                                cur_z++;
                            }
                            break;
                        case NEGATIVE_X: /* Segment in X- direction */
                            if (!ourBlocks.getFlag(cur_x - 1, cur_z)) { /* Right turn? */
                                lineList.add(new int[]{cur_x, cur_z + 1}); /* Finish line */
                                dir = Direction.NEGATIVE_Z;  /* Change direction */
                            } else if (!ourBlocks.getFlag(cur_x - 1, cur_z + 1)) {  /* Straight? */
                                cur_x--;
                            } else {  /* Left turn */
                                lineList.add(new int[]{cur_x, cur_z + 1}); /* Finish line */
                                dir = Direction.POSITIVE_Z;
                                cur_x--;
                                cur_z++;
                            }
                            break;
                        case NEGATIVE_Z: /* Segment in Z- direction */
                            if (!ourBlocks.getFlag(cur_x, cur_z - 1)) { /* Right turn? */
                                lineList.add(new int[]{cur_x, cur_z}); /* Finish line */
                                dir = Direction.POSITIVE_X;  /* Change direction */
                            } else if (!ourBlocks.getFlag(cur_x - 1, cur_z - 1)) {  /* Straight? */
                                cur_z--;
                            } else {  /* Left turn */
                                lineList.add(new int[]{cur_x, cur_z}); /* Finish line */
                                dir = Direction.NEGATIVE_X;
                                cur_x--;
                                cur_z--;
                            }
                            break;
                    }
                }
                /* Build information for specific area */
                String polyId = factionName + "__" + world + "__" + poly_index;
                int sz = lineList.size();
                x = new double[sz];
                z = new double[sz];
                for (int i = 0; i < sz; i++) {
                    int[] line = lineList.get(i);
                    x[i] = (double) line[0] * (double) dynmapFactionsPlugin.getBlockSize();
                    z[i] = (double) line[1] * (double) dynmapFactionsPlugin.getBlockSize();
                }
                /* Find existing one */
                AreaMarker m = dynmapFactionsPlugin.getFactionAreaMarkers().remove(polyId); /* Existing area? */
                if (m == null) {
                    m = dynmapFactionsPlugin.getSet().createAreaMarker(polyId, factionName, false, world, x, z, false);
                    if (m == null) {
                        DynmapFactionsPlugin.info("error adding area marker " + polyId);
                        return;
                    }
                } else {
                    m.setCornerLocations(x, z); /* Replace corner locations */
                    m.setLabel(factionName);   /* Update label */
                }
                m.setDescription(desc); /* Set popup */

                /* Set line and fill properties */
                addStyle(factionName, m);

                /* Add to map */
                newmap.put(polyId, m);
                poly_index++;
            }
        }
    }

    /**
     * Find all contiguous blocks, set in target and clear in source
     */
    private void floodFillTarget(TileFlags source, TileFlags destination, int x, int y) {
        Deque<int[]> stack = new ArrayDeque<>();
        stack.push(new int[]{x, y});

        while (!stack.isEmpty()) {
            int[] next = stack.pop();
            x = next[0];
            y = next[1];
            if (source.getFlag(x, y)) { /* Set in src */
                source.setFlag(x, y, false);   /* Clear source */
                destination.setFlag(x, y, true);   /* Set in destination */
                if (source.getFlag(x + 1, y)) {
                    stack.push(new int[]{x + 1, y});
                }
                if (source.getFlag(x - 1, y)) {
                    stack.push(new int[]{x - 1, y});
                }
                if (source.getFlag(x, y + 1)) {
                    stack.push(new int[]{x, y + 1});
                }
                if (source.getFlag(x, y - 1)) {
                    stack.push(new int[]{x, y - 1});
                }
            }
        }
    }

    private void addStyle(String factionId, AreaMarker areaMarker) {
        AreaStyle as = dynmapFactionsPlugin.getCustomStyle().get(factionId);
        if (as == null) {
            as = dynmapFactionsPlugin.getDefaultStyle();
        }
        int sc = 0xFF0000;
        int fc = 0xFF0000;
        try {
            sc = Integer.parseInt(as.strokeColor().substring(1), 16);
            fc = Integer.parseInt(as.fillColor().substring(1), 16);
        } catch (NumberFormatException ignored) {
        }
        areaMarker.setLineStyle(as.strokeWeight(), as.strokeOpacity(), sc);
        areaMarker.setFillStyle(as.fillOpacity(), fc);
        areaMarker.setBoostFlag(as.boost());
    }

    private MarkerIcon getMarkerIcon(String factionName) {
        AreaStyle as = dynmapFactionsPlugin.getCustomStyle().get(factionName);
        if (as == null) {
            as = dynmapFactionsPlugin.getDefaultStyle();
        }
        return as.homeIcon();
    }

    private boolean isVisible(String id, String worldName) {
        if ((dynmapFactionsPlugin.getVisible() != null) && (!dynmapFactionsPlugin.getVisible().isEmpty())) {
            if ((!dynmapFactionsPlugin.getVisible().contains(id)) && (!dynmapFactionsPlugin.getVisible().contains("world:" + worldName))) {
                return false;
            }
        }
        if ((dynmapFactionsPlugin.getHidden() != null) && (!dynmapFactionsPlugin.getHidden().isEmpty())) {
            return !dynmapFactionsPlugin.getHidden().contains(id) && !dynmapFactionsPlugin.getHidden().contains("world:" + worldName);
        }
        return true;
    }

    private String formatInfoWindow(Faction fact) {
        String output = "<div class=\"regionInfo\">" + infoWindow + "</div>";
        output = output.replace("%regionName%", ChatColor.stripColor(fact.getName()));
        output = output.replace("%description%", ChatColor.stripColor(fact.getDescription() != null ? fact.getDescription() : ""));
        MPlayer adm = fact.getLeader();
        output = output.replace("%playerOwners%", (adm != null) ? adm.getName() : "");
        StringBuilder res = new StringBuilder();
        for (MPlayer r : fact.getMPlayers()) {
            if (!res.isEmpty()) {
                res.append(", ");
            }
            res.append(r.getName());
        }
        output = output.replace("%playerMembers%", res.toString());

        // nation is an alias for regionName
        output = output.replace("%nation%", ChatColor.stripColor(fact.getName()));
        /* Build flags */
        StringBuilder flags = new StringBuilder();
        for (MFlag ff : MFlagColl.get().getAll()) {
            flags.append(ff.getName()).append(": ").append(fact.getFlag(ff)).append("<br/>");
            output = output.replace("%flag." + ff.getName() + "%", fact.getFlag(ff) ? "true" : "false");
        }
        output = output.replace("%flags%", flags.toString());
        return output;
    }

}
